@@ -1,6 +1,6 @@
-import hoistVariables from "babel-helper-hoist-variables";
-import type { NodePath } from "babel-traverse";
-import * as t from "babel-types";
+import hoistVariables from "@babel/helper-hoist-variables";
+import type { NodePath } from "@babel/traverse";
+import * as t from "@babel/types";
 
 const visitor = {
   enter(path, state) {
@@ -15,27 +15,33 @@ const visitor = {
 
   Function(path) {
     path.skip();
-  }
+  },
 };
 
-export default function (path: NodePath, scope = path.scope) {
+export default function(path: NodePath, scope = path.scope) {
   const { node } = path;
-  const container = t.functionExpression(null, [], node.body, node.generator, node.async);
+  const container = t.functionExpression(
+    null,
+    [],
+    node.body,
+    node.generator,
+    node.async,
+  );
 
   let callee = container;
-  let args   = [];
+  let args = [];
 
   // todo: only hoist if necessary
-  hoistVariables(path, (id) => scope.push({ id }));
+  hoistVariables(path, id => scope.push({ id }));
 
   const state = {
     foundThis: false,
-    foundArguments: false
+    foundArguments: false,
   };
 
   path.traverse(visitor, state);
 
-  if (state.foundArguments) {
+  if (state.foundArguments || state.foundThis) {
     callee = t.memberExpression(container, t.identifier("apply"));
     args = [];
 
