@@ -11,4 +11,4 @@ var expected = multiline([
   '}} />;',
 ]);
 
-assert.equal(actual, expected);
+expect(actual).toBe(expected);
