@@ -4,37 +4,11 @@ import Renamer from "./lib/renamer";
 import type NodePath from "../path";
 import traverse from "../index";
 import defaults from "lodash/defaults";
-import * as messages from "babel-messages";
 import Binding from "./binding";
 import globals from "globals";
-import * as t from "babel-types";
+import * as t from "@babel/types";
 import { scope as scopeCache } from "../cache";
 
-// Number of calls to the crawl method to figure out whether we're
-// somewhere inside a call that was trigerred by call. This is meant
-// to be used to figure out whether a warning should be trigerred.
-// See `warnOnFlowBinding`.
-let _crawlCallsCount = 0;
-
-/**
- * To avoid creating a new Scope instance for each traversal, we maintain a cache on the
- * node itself containing all scopes it has been associated with.
- */
-
-function getCache(path, parentScope, self) {
-  const scopes: Array<Scope> = scopeCache.get(path.node) || [];
-
-  for (const scope of scopes) {
-    if (scope.parent === parentScope && scope.path === path) return scope;
-  }
-
-  scopes.push(self);
-
-  if (!scopeCache.has(path.node)) {
-    scopeCache.set(path.node, scopes);
-  }
-}
-
 // Recursively gathers the identifying names of a node.
 function gatherNodeParts(node: Object, parts: Array) {
   if (t.isModuleDeclaration(node)) {
@@ -62,6 +36,12 @@ function gatherNodeParts(node: Object, parts: Array) {
     for (const prop of (node.properties: Array)) {
       gatherNodeParts(prop.key || prop.argument, parts);
     }
+  } else if (t.isPrivateName(node)) {
+    gatherNodeParts(node.id, parts);
+  } else if (t.isThisExpression(node)) {
+    parts.push("this");
+  } else if (t.isSuper(node)) {
+    parts.push("super");
   }
 }
 
@@ -71,7 +51,11 @@ const collectorVisitor = {
   For(path) {
     for (const key of (t.FOR_INIT_KEYS: Array)) {
       const declar = path.get(key);
-      if (declar.isVar()) path.scope.getFunctionParent().registerBinding("var", declar);
+      if (declar.isVar()) {
+        const parentScope =
+          path.scope.getFunctionParent() || path.scope.getProgramParent();
+        parentScope.registerBinding("var", declar);
+      }
     }
   },
 
@@ -80,13 +64,14 @@ const collectorVisitor = {
     if (path.isBlockScoped()) return;
 
     // this will be hit again once we traverse into it after this iteration
-    if (path.isExportDeclaration() && path.get("declaration").isDeclaration()) return;
-
-    // TODO(amasad): remove support for flow as bindings (See warning below).
-    //if (path.isFlow()) return;
+    if (path.isExportDeclaration() && path.get("declaration").isDeclaration()) {
+      return;
+    }
 
     // we've ran into a declaration!
-    path.scope.getFunctionParent().registerDeclaration(path);
+    const parent =
+      path.scope.getFunctionParent() || path.scope.getProgramParent();
+    parent.registerDeclaration(path);
   },
 
   ReferencedIdentifier(path, state) {
@@ -96,7 +81,7 @@ const collectorVisitor = {
   ForXStatement(path, state) {
     const left = path.get("left");
     if (left.isPattern() || left.isIdentifier()) {
-      state.constantViolations.push(left);
+      state.constantViolations.push(path);
     }
   },
 
@@ -112,14 +97,13 @@ const collectorVisitor = {
         if (binding) binding.reference(path);
       } else if (t.isVariableDeclaration(declar)) {
         for (const decl of (declar.declarations: Array<Object>)) {
-          const ids = t.getBindingIdentifiers(decl);
-          for (const name in ids) {
+          for (const name of Object.keys(t.getBindingIdentifiers(decl))) {
             const binding = scope.getBinding(name);
             if (binding) binding.reference(path);
           }
         }
       }
-    }
+    },
   },
 
   LabeledStatement(path) {
@@ -132,12 +116,12 @@ const collectorVisitor = {
   },
 
   UpdateExpression(path, state) {
-    state.constantViolations.push(path.get("argument"));
+    state.constantViolations.push(path);
   },
 
   UnaryExpression(path, state) {
     if (path.node.operator === "delete") {
-      state.constantViolations.push(path.get("argument"));
+      state.constantViolations.push(path);
     }
   },
 
@@ -162,35 +146,33 @@ const collectorVisitor = {
         path.scope.getBlockParent().registerDeclaration(bodyPath);
       }
     }
-  }
+  },
 };
 
 let uid = 0;
 
 export default class Scope {
-
   /**
    * This searches the current "scope" and collects all references/bindings
    * within.
    */
 
-  constructor(path: NodePath, parentScope?: Scope) {
-    if (parentScope && parentScope.block === path.node) {
-      return parentScope;
+  constructor(path: NodePath) {
+    const { node } = path;
+    const cached = scopeCache.get(node);
+    // Sometimes, a scopable path is placed higher in the AST tree.
+    // In these cases, have to create a new Scope.
+    if (cached && cached.path === path) {
+      return cached;
     }
-
-    const cached = getCache(path, parentScope, this);
-    if (cached) return cached;
+    scopeCache.set(node, this);
 
     this.uid = uid++;
-    this.parent = parentScope;
-    this.hub    = path.hub;
 
-    this.parentBlock = path.parent;
-    this.block       = path.node;
-    this.path        = path;
+    this.block = node;
+    this.path = path;
 
-    this.labels      = new Map();
+    this.labels = new Map();
   }
 
   /**
@@ -203,12 +185,20 @@ export default class Scope {
    * Variables available in current context.
    */
 
-  static contextVariables = [
-    "arguments",
-    "undefined",
-    "Infinity",
-    "NaN"
-  ];
+  static contextVariables = ["arguments", "undefined", "Infinity", "NaN"];
+
+  get parent() {
+    const parent = this.path.findParent(p => p.isScope());
+    return parent && parent.scope;
+  }
+
+  get parentBlock() {
+    return this.path.parent;
+  }
+
+  get hub() {
+    return this.path.hub;
+  }
 
   /**
    * Traverse node with current scope and path.
@@ -222,17 +212,17 @@ export default class Scope {
    * Generate a unique identifier and add it to the current scope.
    */
 
-  generateDeclaredUidIdentifier(name: string = "temp") {
+  generateDeclaredUidIdentifier(name?: string) {
     const id = this.generateUidIdentifier(name);
     this.push({ id });
-    return id;
+    return t.cloneNode(id);
   }
 
   /**
    * Generate a unique identifier.
    */
 
-  generateUidIdentifier(name: string = "temp") {
+  generateUidIdentifier(name?: string) {
     return t.identifier(this.generateUid(name));
   }
 
@@ -241,14 +231,22 @@ export default class Scope {
    */
 
   generateUid(name: string = "temp") {
-    name = t.toIdentifier(name).replace(/^_+/, "").replace(/[0-9]+$/g, "");
+    name = t
+      .toIdentifier(name)
+      .replace(/^_+/, "")
+      .replace(/[0-9]+$/g, "");
 
     let uid;
     let i = 0;
     do {
       uid = this._generateUid(name, i);
       i++;
-    } while (this.hasLabel(uid) || this.hasBinding(uid) || this.hasGlobal(uid) || this.hasReference(uid));
+    } while (
+      this.hasLabel(uid) ||
+      this.hasBinding(uid) ||
+      this.hasGlobal(uid) ||
+      this.hasReference(uid)
+    );
 
     const program = this.getProgramParent();
     program.references[uid] = true;
@@ -267,11 +265,7 @@ export default class Scope {
     return `_${id}`;
   }
 
-  /**
-   * Generate a unique identifier based on a node.
-   */
-
-  generateUidIdentifierBasedOnNode(parent: Object, defaultName?: String):  Object {
+  generateUidBasedOnNode(parent: Object, defaultName?: String) {
     let node = parent;
 
     if (t.isAssignmentExpression(parent)) {
@@ -288,7 +282,18 @@ export default class Scope {
     let id = parts.join("$");
     id = id.replace(/^_/, "") || defaultName || "ref";
 
-    return this.generateUidIdentifier(id.slice(0, 20));
+    return this.generateUid(id.slice(0, 20));
+  }
+
+  /**
+   * Generate a unique identifier based on a node.
+   */
+
+  generateUidIdentifierBasedOnNode(
+    parent: Object,
+    defaultName?: String,
+  ): Object {
+    return t.identifier(this.generateUidBasedOnNode(parent, defaultName));
   }
 
   /**
@@ -327,7 +332,10 @@ export default class Scope {
       return null;
     } else {
       const id = this.generateUidIdentifierBasedOnNode(node);
-      if (!dontPush) this.push({ id });
+      if (!dontPush) {
+        this.push({ id });
+        return t.cloneNode(id);
+      }
       return id;
     }
   }
@@ -336,17 +344,25 @@ export default class Scope {
     // ignore parameters
     if (kind === "param") return;
 
-    // ignore hoisted functions if there's also a local let
-    if (kind === "hoisted" && local.kind === "let") return;
+    // Ignore existing binding if it's the name of the current function or
+    // class expression
+    if (local.kind === "local") return;
 
     const duplicate =
       // don't allow duplicate bindings to exist alongside
-      kind === "let" || local.kind === "let" || local.kind === "const" || local.kind === "module" ||
+      kind === "let" ||
+      local.kind === "let" ||
+      local.kind === "const" ||
+      local.kind === "module" ||
       // don't allow a local of param with a kind of let
-      local.kind === "param" && (kind === "let" || kind === "const");
+      (local.kind === "param" && (kind === "let" || kind === "const"));
 
     if (duplicate) {
-      throw this.hub.file.buildCodeFrameError(id, messages.get("scopeDuplicateDeclaration", name), TypeError);
+      throw this.hub.buildError(
+        id,
+        `Duplicate declaration "${name}"`,
+        TypeError,
+      );
     }
   }
 
@@ -371,25 +387,25 @@ export default class Scope {
     let scope = this;
     do {
       console.log("#", scope.block.type);
-      for (const name in scope.bindings) {
+      for (const name of Object.keys(scope.bindings)) {
         const binding = scope.bindings[name];
         console.log(" -", name, {
           constant: binding.constant,
           references: binding.references,
           violations: binding.constantViolations.length,
-          kind: binding.kind
+          kind: binding.kind,
         });
       }
-    } while (scope = scope.parent);
+    } while ((scope = scope.parent));
     console.log(sep);
   }
 
   toArray(node: Object, i?: number) {
-    const file = this.hub.file;
-
     if (t.isIdentifier(node)) {
       const binding = this.getBinding(node.name);
-      if (binding && binding.constant && binding.path.isGenericType("Array")) return node;
+      if (binding && binding.constant && binding.path.isGenericType("Array")) {
+        return node;
+      }
     }
 
     if (t.isArrayExpression(node)) {
@@ -402,26 +418,32 @@ export default class Scope {
           t.memberExpression(
             t.memberExpression(
               t.identifier("Array"),
-              t.identifier("prototype")
+              t.identifier("prototype"),
             ),
-            t.identifier("slice")
+            t.identifier("slice"),
           ),
-          t.identifier("call")
+          t.identifier("call"),
         ),
-        [node]
+        [node],
       );
     }
 
-    let helperName = "toArray";
+    let helperName;
     const args = [node];
     if (i === true) {
+      // Used in array-spread to create an array.
       helperName = "toConsumableArray";
     } else if (i) {
       args.push(t.numericLiteral(i));
+
+      // Used in array-rest to create an array from a subset of an iterable.
       helperName = "slicedToArray";
-      // TODO if (this.hub.file.isLoose("es6.forOf")) helperName += "-loose";
+      // TODO if (this.hub.isLoose("es6.forOf")) helperName += "-loose";
+    } else {
+      // Used in array-rest to create an array
+      helperName = "toArray";
     }
-    return t.callExpression(file.addHelper(helperName), args);
+    return t.callExpression(this.hub.addHelper(helperName), args);
   }
 
   hasLabel(name: string) {
@@ -455,7 +477,11 @@ export default class Scope {
       }
     } else if (path.isExportDeclaration()) {
       const declar = path.get("declaration");
-      if (declar.isClassDeclaration() || declar.isFunctionDeclaration() || declar.isVariableDeclaration()) {
+      if (
+        declar.isClassDeclaration() ||
+        declar.isFunctionDeclaration() ||
+        declar.isVariableDeclaration()
+      ) {
         this.registerDeclaration(declar);
       }
     } else {
@@ -467,13 +493,14 @@ export default class Scope {
     if (this.hasBinding("undefined")) {
       return t.unaryExpression("void", t.numericLiteral(0), true);
     } else {
+      // eslint-disable-next-line @babel/development/no-undefined-identifier
       return t.identifier("undefined");
     }
   }
 
   registerConstantViolation(path: NodePath) {
     const ids = path.getBindingIdentifiers();
-    for (const name in ids) {
+    for (const name of Object.keys(ids)) {
       const binding = this.getBinding(name);
       if (binding) binding.reassign(path);
     }
@@ -491,11 +518,12 @@ export default class Scope {
     }
 
     const parent = this.getProgramParent();
-    const ids = path.getBindingIdentifiers(true);
+    const ids = path.getOuterBindingIdentifiers(true);
 
-    for (const name in ids) {
+    for (const name of Object.keys(ids)) {
       for (const id of (ids[name]: Array<Object>)) {
-        let local = this.getOwnBinding(name);
+        const local = this.getOwnBinding(name);
+
         if (local) {
           // same identifier so continue safely as we're likely trying to register it
           // multiple times
@@ -504,20 +532,19 @@ export default class Scope {
           this.checkBlockScopedCollisions(local, kind, name, id);
         }
 
-        // It's erroneous that we currently consider flow a binding, however, we can't
-        // remove it because people might be depending on it. See warning section
-        // in `warnOnFlowBinding`.
-        if (local && local.path.isFlow()) local = null;
-
         parent.references[name] = true;
 
-        this.bindings[name] = new Binding({
-          identifier: id,
-          existing:   local,
-          scope:      this,
-          path:       bindingPath,
-          kind:       kind
-        });
+        // A redeclaration of an existing variable is a modification
+        if (local) {
+          this.registerConstantViolation(bindingPath);
+        } else {
+          this.bindings[name] = new Binding({
+            identifier: id,
+            scope: this,
+            path: bindingPath,
+            kind: kind,
+          });
+        }
       }
     }
   }
@@ -531,7 +558,7 @@ export default class Scope {
 
     do {
       if (scope.uids[name]) return true;
-    } while (scope = scope.parent);
+    } while ((scope = scope.parent));
 
     return false;
   }
@@ -541,7 +568,7 @@ export default class Scope {
 
     do {
       if (scope.globals[name]) return true;
-    } while (scope = scope.parent);
+    } while ((scope = scope.parent));
 
     return false;
   }
@@ -551,7 +578,7 @@ export default class Scope {
 
     do {
       if (scope.references[name]) return true;
-    } while (scope = scope.parent);
+    } while ((scope = scope.parent));
 
     return false;
   }
@@ -563,7 +590,9 @@ export default class Scope {
       if (constantsOnly) return binding.constant;
       return true;
     } else if (t.isClass(node)) {
-      if (node.superClass && !this.isPure(node.superClass, constantsOnly)) return false;
+      if (node.superClass && !this.isPure(node.superClass, constantsOnly)) {
+        return false;
+      }
       return this.isPure(node.body, constantsOnly);
     } else if (t.isClassBody(node)) {
       for (const method of node.body) {
@@ -571,7 +600,10 @@ export default class Scope {
       }
       return true;
     } else if (t.isBinary(node)) {
-      return this.isPure(node.left, constantsOnly) && this.isPure(node.right, constantsOnly);
+      return (
+        this.isPure(node.left, constantsOnly) &&
+        this.isPure(node.right, constantsOnly)
+      );
     } else if (t.isArrayExpression(node)) {
       for (const elem of (node.elements: Array<Object>)) {
         if (!this.isPure(elem, constantsOnly)) return false;
@@ -586,11 +618,22 @@ export default class Scope {
       if (node.computed && !this.isPure(node.key, constantsOnly)) return false;
       if (node.kind === "get" || node.kind === "set") return false;
       return true;
-    } else if (t.isClassProperty(node) || t.isObjectProperty(node)) {
+    } else if (t.isProperty(node)) {
       if (node.computed && !this.isPure(node.key, constantsOnly)) return false;
       return this.isPure(node.value, constantsOnly);
     } else if (t.isUnaryExpression(node)) {
       return this.isPure(node.argument, constantsOnly);
+    } else if (t.isTaggedTemplateExpression(node)) {
+      return (
+        t.matchesPattern(node.tag, "String.raw") &&
+        !this.hasBinding("String", true) &&
+        this.isPure(node.quasi, constantsOnly)
+      );
+    } else if (t.isTemplateLiteral(node)) {
+      for (const expression of (node.expressions: Array<Object>)) {
+        if (!this.isPure(expression, constantsOnly)) return false;
+      }
+      return true;
     } else {
       return t.isPureish(node);
     }
@@ -601,7 +644,7 @@ export default class Scope {
    */
 
   setData(key, val) {
-    return this.data[key] = val;
+    return (this.data[key] = val);
   }
 
   /**
@@ -613,7 +656,7 @@ export default class Scope {
     do {
       const data = scope.data[key];
       if (data != null) return data;
-    } while (scope = scope.parent);
+    } while ((scope = scope.parent));
   }
 
   /**
@@ -626,7 +669,7 @@ export default class Scope {
     do {
       const data = scope.data[key];
       if (data != null) scope.data[key] = null;
-    } while (scope = scope.parent);
+    } while ((scope = scope.parent));
   }
 
   init() {
@@ -634,21 +677,13 @@ export default class Scope {
   }
 
   crawl() {
-    _crawlCallsCount++;
-    this._crawl();
-    _crawlCallsCount--;
-  }
-
-  _crawl() {
     const path = this.path;
 
-    //
-
     this.references = Object.create(null);
-    this.bindings   = Object.create(null);
-    this.globals    = Object.create(null);
-    this.uids       = Object.create(null);
-    this.data       = Object.create(null);
+    this.bindings = Object.create(null);
+    this.globals = Object.create(null);
+    this.uids = Object.create(null);
+    this.data = Object.create(null);
 
     // ForStatement - left, init
 
@@ -710,10 +745,10 @@ export default class Scope {
       // register undeclared bindings as globals
       const ids = path.getBindingIdentifiers();
       let programParent;
-      for (const name in ids) {
+      for (const name of Object.keys(ids)) {
         if (path.scope.getBinding(name)) continue;
 
-        programParent = programParent ||  path.scope.getProgramParent();
+        programParent = programParent || path.scope.getProgramParent();
         programParent.addGlobal(ids[name]);
       }
 
@@ -738,11 +773,11 @@ export default class Scope {
   }
 
   push(opts: {
-    id: Object;
-    init: ?Object;
-    unique: ?boolean;
-    _blockHoist: ?number;
-    kind: "var" | "let";
+    id: Object,
+    init: ?Object,
+    unique: ?boolean,
+    _blockHoist: ?number,
+    kind: "var" | "let",
   }) {
     let path = this.path;
 
@@ -751,24 +786,23 @@ export default class Scope {
     }
 
     if (path.isSwitchStatement()) {
-      path = this.getFunctionParent().path;
+      path = (this.getFunctionParent() || this.getProgramParent()).path;
     }
 
     if (path.isLoop() || path.isCatchClause() || path.isFunction()) {
-      t.ensureBlock(path.node);
+      path.ensureBlock();
       path = path.get("body");
     }
 
     const unique = opts.unique;
-    const kind   = opts.kind || "var";
+    const kind = opts.kind || "var";
     const blockHoist = opts._blockHoist == null ? 2 : opts._blockHoist;
 
     const dataKey = `declaration:${kind}:${blockHoist}`;
-    let declarPath  = !unique && path.getData(dataKey);
+    let declarPath = !unique && path.getData(dataKey);
 
     if (!declarPath) {
       const declar = t.variableDeclaration(kind, []);
-      declar._generated = true;
       declar._blockHoist = blockHoist;
 
       [declarPath] = path.unshiftContainer("body", [declar]);
@@ -790,13 +824,12 @@ export default class Scope {
       if (scope.path.isProgram()) {
         return scope;
       }
-    } while (scope = scope.parent);
-    throw new Error("We couldn't find a Function or Program...");
+    } while ((scope = scope.parent));
+    throw new Error("Couldn't find a Program");
   }
 
   /**
-   * Walk up the scope tree until we hit either a Function or reach the
-   * very top and hit Program.
+   * Walk up the scope tree until we hit either a Function or return null.
    */
 
   getFunctionParent() {
@@ -805,8 +838,8 @@ export default class Scope {
       if (scope.path.isFunctionParent()) {
         return scope;
       }
-    } while (scope = scope.parent);
-    throw new Error("We couldn't find a Function or Program...");
+    } while ((scope = scope.parent));
+    return null;
   }
 
   /**
@@ -820,8 +853,10 @@ export default class Scope {
       if (scope.path.isBlockParent()) {
         return scope;
       }
-    } while (scope = scope.parent);
-    throw new Error("We couldn't find a BlockStatement, For, Switch, Function, Loop or Program...");
+    } while ((scope = scope.parent));
+    throw new Error(
+      "We couldn't find a BlockStatement, For, Switch, Function, Loop or Program...",
+    );
   }
 
   /**
@@ -850,7 +885,7 @@ export default class Scope {
     for (const kind of (arguments: Array)) {
       let scope = this;
       do {
-        for (const name in scope.bindings) {
+        for (const name of Object.keys(scope.bindings)) {
           const binding = scope.bindings[name];
           if (binding.kind === kind) ids[name] = binding;
         }
@@ -865,28 +900,17 @@ export default class Scope {
     return this.getBindingIdentifier(name) === node;
   }
 
-  warnOnFlowBinding(binding) {
-    if (_crawlCallsCount === 0 && binding && binding.path.isFlow()) {
-      console.warn(`
-        You or one of the Babel plugins you are using are using Flow declarations as bindings.
-        Support for this will be removed in version 7. To find out the caller, grep for this
-        message and change it to a \`console.trace()\`.
-      `);
-    }
-    return binding;
-  }
-
   getBinding(name: string) {
     let scope = this;
 
     do {
       const binding = scope.getOwnBinding(name);
-      if (binding) return this.warnOnFlowBinding(binding);
-    } while (scope = scope.parent);
+      if (binding) return binding;
+    } while ((scope = scope.parent));
   }
 
   getOwnBinding(name: string) {
-    return this.warnOnFlowBinding(this.bindings[name]);
+    return this.bindings[name];
   }
 
   getBindingIdentifier(name: string) {
@@ -947,6 +971,6 @@ export default class Scope {
       if (scope.uids[name]) {
         scope.uids[name] = false;
       }
-    } while (scope = scope.parent);
+    } while ((scope = scope.parent));
   }
 }
